# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Optional
from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..ascii import Connection, Device, Axis, AxisGroup
from .illuminator import Illuminator
from .filter_changer import FilterChanger
from .objective_changer import ObjectiveChanger
from ..dto.microscopy.microscope_config import MicroscopeConfig


class Microscope:
    """
    Represent a microscope.
    Parts of the microscope may or may not be instantiated depending on the configuration.
    Requires at least Firmware 7.34.
    """

    @property
    def connection(self) -> Connection:
        """
        Connection of the microscope.
        """
        return self._connection

    @property
    def illuminator(self) -> Optional[Illuminator]:
        """
        The illuminator.
        """
        return self._illuminator

    @property
    def focus_axis(self) -> Optional[Axis]:
        """
        The focus axis.
        """
        return self._focus_axis

    @property
    def x_axis(self) -> Optional[Axis]:
        """
        The X axis.
        """
        return self._x_axis

    @property
    def y_axis(self) -> Optional[Axis]:
        """
        The Y axis.
        """
        return self._y_axis

    @property
    def plate(self) -> Optional[AxisGroup]:
        """
        Axis group consisting of X and Y axes representing the plate of the microscope.
        """
        return self._plate

    @property
    def objective_changer(self) -> Optional[ObjectiveChanger]:
        """
        The objective changer.
        """
        return self._objective_changer

    @property
    def filter_changer(self) -> Optional[FilterChanger]:
        """
        The filter changer.
        """
        return self._filter_changer

    def __init__(self, connection: Connection, config: MicroscopeConfig):
        """
        Creates instance of `Microscope` from the given config.
        Parts are instantiated depending on device addresses in the config.
        """
        self._connection = connection
        self._config = MicroscopeConfig.from_binary(MicroscopeConfig.to_binary(config))
        self._illuminator = Illuminator(Device(connection, config.illuminator)) if config.illuminator else None
        self._focus_axis = Axis(Device(connection, config.focus_axis.device), config.focus_axis.axis)\
            if config.focus_axis and config.focus_axis.device else None
        self._x_axis = Axis(Device(connection, config.x_axis.device), config.x_axis.axis)\
            if config.x_axis and config.x_axis.device else None
        self._y_axis = Axis(Device(connection, config.y_axis.device), config.y_axis.axis)\
            if config.y_axis and config.y_axis.device else None
        self._plate = AxisGroup([self._x_axis, self._y_axis])\
            if self._x_axis is not None and self._y_axis is not None else None
        self._objective_changer = ObjectiveChanger(Device(connection, config.objective_changer), self._focus_axis)\
            if config.objective_changer and self._focus_axis else None
        self._filter_changer = FilterChanger(Device(connection, config.filter_changer))\
            if config.filter_changer else None

    @staticmethod
    def find(
            connection: Connection
    ) -> 'Microscope':
        """
        Finds a microscope on a connection.

        Args:
            connection: Connection on which to detect the microscope.

        Returns:
            New instance of microscope.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=connection.interface_id,
        )
        response = call(
            "microscope/detect",
            request,
            dto.MicroscopeConfigResponse.from_binary)
        return Microscope(connection, response.config)

    @staticmethod
    async def find_async(
            connection: Connection
    ) -> 'Microscope':
        """
        Finds a microscope on a connection.

        Args:
            connection: Connection on which to detect the microscope.

        Returns:
            New instance of microscope.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=connection.interface_id,
        )
        response = await call_async(
            "microscope/detect",
            request,
            dto.MicroscopeConfigResponse.from_binary)
        return Microscope(connection, response.config)

    def initialize(
            self,
            force: bool = False
    ) -> None:
        """
        Initializes the microscope.
        Homes all axes, filter changer, and objective changer if they require it.

        Args:
            force: Forces all devices to home even when not required.
        """
        request = dto.MicroscopeInitRequest(
            interface_id=self.connection.interface_id,
            config=self._config,
            force=force,
        )
        call("microscope/initialize", request)

    async def initialize_async(
            self,
            force: bool = False
    ) -> None:
        """
        Initializes the microscope.
        Homes all axes, filter changer, and objective changer if they require it.

        Args:
            force: Forces all devices to home even when not required.
        """
        request = dto.MicroscopeInitRequest(
            interface_id=self.connection.interface_id,
            config=self._config,
            force=force,
        )
        await call_async("microscope/initialize", request)

    def is_initialized(
            self
    ) -> bool:
        """
        Checks whether the microscope is initialized.

        Returns:
            True, when the microscope is initialized. False, otherwise.
        """
        request = dto.MicroscopeEmptyRequest(
            interface_id=self.connection.interface_id,
            config=self._config,
        )
        response = call(
            "microscope/is_initialized",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_initialized_async(
            self
    ) -> bool:
        """
        Checks whether the microscope is initialized.

        Returns:
            True, when the microscope is initialized. False, otherwise.
        """
        request = dto.MicroscopeEmptyRequest(
            interface_id=self.connection.interface_id,
            config=self._config,
        )
        response = await call_async(
            "microscope/is_initialized",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the microscope.

        Returns:
            A string that represents the microscope.
        """
        request = dto.MicroscopeEmptyRequest(
            interface_id=self.connection.interface_id,
            config=self._config,
        )
        response = call_sync(
            "microscope/to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..ascii import DeviceIO, Connection, Device
from .illuminator_channel import IlluminatorChannel


class Illuminator:
    """
    Use to manage an LED controller.
    Requires at least Firmware 7.09.
    """

    @property
    def device(self) -> Device:
        """
        The base device of this illuminator.
        """
        return self._device

    @property
    def io(self) -> DeviceIO:
        """
        I/O channels of this device.
        """
        return self._io

    def __init__(self, device: Device):
        """
        Creates instance of `Illuminator` based on the given device.
        If the device is identified, this constructor will ensure it is an illuminator.
        """
        self._device = device
        self._io = DeviceIO(device)
        self.__verify_is_illuminator()

    def get_channel(
            self,
            channel_number: int
    ) -> IlluminatorChannel:
        """
        Gets an IlluminatorChannel class instance that allows control of a particular channel.
        Channels are numbered from 1.

        Args:
            channel_number: Number of channel to control.

        Returns:
            Illuminator channel instance.
        """
        if channel_number <= 0:
            raise ValueError('Invalid value; channels are numbered from 1.')

        return IlluminatorChannel(self, channel_number)

    def __verify_is_illuminator(
            self
    ) -> None:
        """
        Checks if this is an illuminator or some other type of device and throws an error if it is not.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        call_sync("illuminator/verify", request)

    @staticmethod
    def find(
            connection: Connection,
            device_address: int = 0
    ) -> 'Illuminator':
        """
        Finds an illuminator on a connection.
        In case of conflict, specify the optional device address.

        Args:
            connection: Connection on which to detect the illuminator.
            device_address: Optional device address of the illuminator.

        Returns:
            New instance of illuminator.
        """
        request = dto.FindDeviceRequest(
            interface_id=connection.interface_id,
            device_address=device_address,
        )
        response = call(
            "illuminator/detect",
            request,
            dto.FindDeviceResponse.from_binary)
        return Illuminator(Device(connection, response.address))

    @staticmethod
    async def find_async(
            connection: Connection,
            device_address: int = 0
    ) -> 'Illuminator':
        """
        Finds an illuminator on a connection.
        In case of conflict, specify the optional device address.

        Args:
            connection: Connection on which to detect the illuminator.
            device_address: Optional device address of the illuminator.

        Returns:
            New instance of illuminator.
        """
        request = dto.FindDeviceRequest(
            interface_id=connection.interface_id,
            device_address=device_address,
        )
        response = await call_async(
            "illuminator/detect",
            request,
            dto.FindDeviceResponse.from_binary)
        return Illuminator(Device(connection, response.address))

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call_sync(
            "device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .filter_changer import FilterChanger as FilterChanger
from .illuminator import Illuminator as Illuminator
from .illuminator_channel import IlluminatorChannel as IlluminatorChannel
from .microscope import Microscope as Microscope
from .objective_changer import ObjectiveChanger as ObjectiveChanger
from ..dto.microscopy.microscope_config import MicroscopeConfig as MicroscopeConfig



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class Message:
    """
    A message in the Binary protocol.
    """

    device_address: int
    """
    Number of the device that sent or should receive the message.
    """

    command: int
    """
    The warning flag contains the highest priority warning currently active for the device or axis.
    """

    data: int
    """
    Data payload of the message, if applicable, or zero otherwise.
    """

    @staticmethod
    def zero_values() -> 'Message':
        return Message(
            device_address=0,
            command=0,
            data=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'Message':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return Message.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'deviceAddress': int(self.device_address),
            'command': int(self.command),
            'data': int(self.data),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'Message':
        return Message(
            device_address=data.get('deviceAddress'),  # type: ignore
            command=data.get('command'),  # type: ignore
            data=data.get('data'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.device_address is None:
            raise ValueError(f'Property "DeviceAddress" of "Message" is None.')

        if not isinstance(self.device_address, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "DeviceAddress" of "Message" is not a number.')

        if int(self.device_address) != self.device_address:
            raise ValueError(f'Property "DeviceAddress" of "Message" is not integer value.')

        if self.command is None:
            raise ValueError(f'Property "Command" of "Message" is None.')

        if not isinstance(self.command, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Command" of "Message" is not a number.')

        if int(self.command) != self.command:
            raise ValueError(f'Property "Command" of "Message" is not integer value.')

        if self.data is None:
            raise ValueError(f'Property "Data" of "Message" is None.')

        if not isinstance(self.data, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Data" of "Message" is not a number.')

        if int(self.data) != self.data:
            raise ValueError(f'Property "Data" of "Message" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from ..ascii.digital_output_action import DigitalOutputAction
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class StreamSetAllDigitalOutputsScheduleRequest:

    interface_id: int = 0

    device: int = 0

    stream_id: int = 0

    pvt: bool = False

    values: List[DigitalOutputAction] = field(default_factory=list)

    future_values: List[DigitalOutputAction] = field(default_factory=list)

    delay: float = 0

    unit: UnitsAndLiterals = Units.NATIVE

    @staticmethod
    def zero_values() -> 'StreamSetAllDigitalOutputsScheduleRequest':
        return StreamSetAllDigitalOutputsScheduleRequest(
            interface_id=0,
            device=0,
            stream_id=0,
            pvt=False,
            values=[],
            future_values=[],
            delay=0,
            unit=Units.NATIVE,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'StreamSetAllDigitalOutputsScheduleRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return StreamSetAllDigitalOutputsScheduleRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'streamId': int(self.stream_id),
            'pvt': bool(self.pvt),
            'values': [item.value for item in self.values] if self.values is not None else [],
            'futureValues': [item.value for item in self.future_values] if self.future_values is not None else [],
            'delay': float(self.delay),
            'unit': units_from_literals(self.unit).value,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'StreamSetAllDigitalOutputsScheduleRequest':
        return StreamSetAllDigitalOutputsScheduleRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            stream_id=data.get('streamId'),  # type: ignore
            pvt=data.get('pvt'),  # type: ignore
            values=[DigitalOutputAction(item) for item in data.get('values')],  # type: ignore
            future_values=[DigitalOutputAction(item) for item in data.get('futureValues')],  # type: ignore
            delay=data.get('delay'),  # type: ignore
            unit=Units(data.get('unit')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "StreamSetAllDigitalOutputsScheduleRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "StreamSetAllDigitalOutputsScheduleRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "StreamSetAllDigitalOutputsScheduleRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "StreamSetAllDigitalOutputsScheduleRequest" is not integer value.')

        if self.stream_id is None:
            raise ValueError(f'Property "StreamId" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

        if not isinstance(self.stream_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "StreamId" of "StreamSetAllDigitalOutputsScheduleRequest" is not a number.')

        if int(self.stream_id) != self.stream_id:
            raise ValueError(f'Property "StreamId" of "StreamSetAllDigitalOutputsScheduleRequest" is not integer value.')

        if self.values is not None:
            if not isinstance(self.values, Iterable):
                raise ValueError('Property "Values" of "StreamSetAllDigitalOutputsScheduleRequest" is not iterable.')

            for i, values_item in enumerate(self.values):
                if values_item is None:
                    raise ValueError(f'Item {i} in property "Values" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

                if not isinstance(values_item, DigitalOutputAction):
                    raise ValueError(f'Item {i} in property "Values" of "StreamSetAllDigitalOutputsScheduleRequest" is not an instance of "DigitalOutputAction".')

        if self.future_values is not None:
            if not isinstance(self.future_values, Iterable):
                raise ValueError('Property "FutureValues" of "StreamSetAllDigitalOutputsScheduleRequest" is not iterable.')

            for i, future_values_item in enumerate(self.future_values):
                if future_values_item is None:
                    raise ValueError(f'Item {i} in property "FutureValues" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

                if not isinstance(future_values_item, DigitalOutputAction):
                    raise ValueError(f'Item {i} in property "FutureValues" of "StreamSetAllDigitalOutputsScheduleRequest" is not an instance of "DigitalOutputAction".')

        if self.delay is None:
            raise ValueError(f'Property "Delay" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

        if not isinstance(self.delay, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Delay" of "StreamSetAllDigitalOutputsScheduleRequest" is not a number.')

        if self.unit is None:
            raise ValueError(f'Property "Unit" of "StreamSetAllDigitalOutputsScheduleRequest" is None.')

        if not isinstance(self.unit, (Units, str)):
            raise ValueError(f'Property "Unit" of "StreamSetAllDigitalOutputsScheduleRequest" is not Units.')



# This file is generated. Do not modify by hand.
from enum import Enum


class IoPortType(Enum):
    """
    Kind of I/O pin to use.
    """

    NONE = 0
    ANALOG_INPUT = 1
    ANALOG_OUTPUT = 2
    DIGITAL_INPUT = 3
    DIGITAL_OUTPUT = 4



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class InvalidPvtPoint:
    """
    Contains invalid PVT points for PvtExecutionException.
    """

    index: int
    """
    Index of the point numbered from the last submitted point.
    """

    point: str
    """
    The textual representation of the point.
    """

    @staticmethod
    def zero_values() -> 'InvalidPvtPoint':
        return InvalidPvtPoint(
            index=0,
            point="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'InvalidPvtPoint':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return InvalidPvtPoint.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'index': int(self.index),
            'point': str(self.point or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'InvalidPvtPoint':
        return InvalidPvtPoint(
            index=data.get('index'),  # type: ignore
            point=data.get('point'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.index is None:
            raise ValueError(f'Property "Index" of "InvalidPvtPoint" is None.')

        if not isinstance(self.index, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Index" of "InvalidPvtPoint" is not a number.')

        if int(self.index) != self.index:
            raise ValueError(f'Property "Index" of "InvalidPvtPoint" is not integer value.')

        if self.point is not None:
            if not isinstance(self.point, str):
                raise ValueError(f'Property "Point" of "InvalidPvtPoint" is not a string.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from ..ascii.conversion_factor import ConversionFactor


@dataclass
class DeviceSetUnitConversionsRequest:

    interface_id: int = 0

    device: int = 0

    axis: int = 0

    key: str = ""

    conversions: List[ConversionFactor] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'DeviceSetUnitConversionsRequest':
        return DeviceSetUnitConversionsRequest(
            interface_id=0,
            device=0,
            axis=0,
            key="",
            conversions=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'DeviceSetUnitConversionsRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return DeviceSetUnitConversionsRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'axis': int(self.axis),
            'key': str(self.key or ''),
            'conversions': [item.to_dict() for item in self.conversions] if self.conversions is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'DeviceSetUnitConversionsRequest':
        return DeviceSetUnitConversionsRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            axis=data.get('axis'),  # type: ignore
            key=data.get('key'),  # type: ignore
            conversions=[ConversionFactor.from_dict(item) for item in data.get('conversions')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "DeviceSetUnitConversionsRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "DeviceSetUnitConversionsRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "DeviceSetUnitConversionsRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "DeviceSetUnitConversionsRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "DeviceSetUnitConversionsRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "DeviceSetUnitConversionsRequest" is not integer value.')

        if self.axis is None:
            raise ValueError(f'Property "Axis" of "DeviceSetUnitConversionsRequest" is None.')

        if not isinstance(self.axis, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis" of "DeviceSetUnitConversionsRequest" is not a number.')

        if int(self.axis) != self.axis:
            raise ValueError(f'Property "Axis" of "DeviceSetUnitConversionsRequest" is not integer value.')

        if self.key is not None:
            if not isinstance(self.key, str):
                raise ValueError(f'Property "Key" of "DeviceSetUnitConversionsRequest" is not a string.')

        if self.conversions is not None:
            if not isinstance(self.conversions, Iterable):
                raise ValueError('Property "Conversions" of "DeviceSetUnitConversionsRequest" is not iterable.')

            for i, conversions_item in enumerate(self.conversions):
                if conversions_item is None:
                    raise ValueError(f'Item {i} in property "Conversions" of "DeviceSetUnitConversionsRequest" is None.')

                if not isinstance(conversions_item, ConversionFactor):
                    raise ValueError(f'Item {i} in property "Conversions" of "DeviceSetUnitConversionsRequest" is not an instance of "ConversionFactor".')

                conversions_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from .stream_segment_type import StreamSegmentType
from ..rotation_direction import RotationDirection
from ..measurement import Measurement


@dataclass
class StreamCircleRequest:

    interface_id: int = 0

    device: int = 0

    stream_id: int = 0

    pvt: bool = False

    type: StreamSegmentType = next(first for first in StreamSegmentType)

    rotation_direction: RotationDirection = next(first for first in RotationDirection)

    center_x: Measurement = field(default_factory=Measurement.zero_values)

    center_y: Measurement = field(default_factory=Measurement.zero_values)

    target_axes_indices: List[int] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'StreamCircleRequest':
        return StreamCircleRequest(
            interface_id=0,
            device=0,
            stream_id=0,
            pvt=False,
            type=next(first for first in StreamSegmentType),
            rotation_direction=next(first for first in RotationDirection),
            center_x=Measurement.zero_values(),
            center_y=Measurement.zero_values(),
            target_axes_indices=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'StreamCircleRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return StreamCircleRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'streamId': int(self.stream_id),
            'pvt': bool(self.pvt),
            'type': self.type.value,
            'rotationDirection': self.rotation_direction.value,
            'centerX': self.center_x.to_dict(),
            'centerY': self.center_y.to_dict(),
            'targetAxesIndices': [int(item) for item in self.target_axes_indices] if self.target_axes_indices is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'StreamCircleRequest':
        return StreamCircleRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            stream_id=data.get('streamId'),  # type: ignore
            pvt=data.get('pvt'),  # type: ignore
            type=StreamSegmentType(data.get('type')),  # type: ignore
            rotation_direction=RotationDirection(data.get('rotationDirection')),  # type: ignore
            center_x=Measurement.from_dict(data.get('centerX')),  # type: ignore
            center_y=Measurement.from_dict(data.get('centerY')),  # type: ignore
            target_axes_indices=data.get('targetAxesIndices'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "StreamCircleRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "StreamCircleRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "StreamCircleRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "StreamCircleRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "StreamCircleRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "StreamCircleRequest" is not integer value.')

        if self.stream_id is None:
            raise ValueError(f'Property "StreamId" of "StreamCircleRequest" is None.')

        if not isinstance(self.stream_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "StreamId" of "StreamCircleRequest" is not a number.')

        if int(self.stream_id) != self.stream_id:
            raise ValueError(f'Property "StreamId" of "StreamCircleRequest" is not integer value.')

        if self.type is None:
            raise ValueError(f'Property "Type" of "StreamCircleRequest" is None.')

        if not isinstance(self.type, StreamSegmentType):
            raise ValueError(f'Property "Type" of "StreamCircleRequest" is not an instance of "StreamSegmentType".')

        if self.rotation_direction is None:
            raise ValueError(f'Property "RotationDirection" of "StreamCircleRequest" is None.')

        if not isinstance(self.rotation_direction, RotationDirection):
            raise ValueError(f'Property "RotationDirection" of "StreamCircleRequest" is not an instance of "RotationDirection".')

        if self.center_x is None:
            raise ValueError(f'Property "CenterX" of "StreamCircleRequest" is None.')

        if not isinstance(self.center_x, Measurement):
            raise ValueError(f'Property "CenterX" of "StreamCircleRequest" is not an instance of "Measurement".')

        self.center_x.validate()

        if self.center_y is None:
            raise ValueError(f'Property "CenterY" of "StreamCircleRequest" is None.')

        if not isinstance(self.center_y, Measurement):
            raise ValueError(f'Property "CenterY" of "StreamCircleRequest" is not an instance of "Measurement".')

        self.center_y.validate()

        if self.target_axes_indices is not None:
            if not isinstance(self.target_axes_indices, Iterable):
                raise ValueError('Property "TargetAxesIndices" of "StreamCircleRequest" is not iterable.')

            for i, target_axes_indices_item in enumerate(self.target_axes_indices):
                if target_axes_indices_item is None:
                    raise ValueError(f'Item {i} in property "TargetAxesIndices" of "StreamCircleRequest" is None.')

                if not isinstance(target_axes_indices_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "TargetAxesIndices" of "StreamCircleRequest" is not a number.')

                if int(target_axes_indices_item) != target_axes_indices_item:
                    raise ValueError(f'Item {i} in property "TargetAxesIndices" of "StreamCircleRequest" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .binary_settings import BinarySettings as BinarySettings
from .command_code import CommandCode as CommandCode
from .device_identity import DeviceIdentity as DeviceIdentity
from .device_type import DeviceType as DeviceType
from .error_code import ErrorCode as ErrorCode
from .message import Message as Message
from .reply_code import ReplyCode as ReplyCode
from .reply_only_event import ReplyOnlyEvent as ReplyOnlyEvent
from .unknown_response_event import UnknownResponseEvent as UnknownResponseEvent



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson
from .set_peripheral_state_exception_data import SetPeripheralStateExceptionData


@dataclass
class SetDeviceStateExceptionData:
    """
    Contains additional data for a SetDeviceStateFailedException.
    """

    settings: List[str]
    """
    A list of settings which could not be set.
    """

    stream_buffers: List[str]
    """
    The reason the stream buffers could not be set.
    """

    pvt_buffers: List[str]
    """
    The reason the pvt buffers could not be set.
    """

    triggers: List[str]
    """
    The reason the triggers could not be set.
    """

    servo_tuning: str
    """
    The reason servo tuning could not be set.
    """

    stored_positions: List[str]
    """
    The reasons stored positions could not be set.
    """

    storage: List[str]
    """
    The reasons storage could not be set.
    """

    peripherals: List[SetPeripheralStateExceptionData]
    """
    Errors for any peripherals that could not be set.
    """

    @staticmethod
    def zero_values() -> 'SetDeviceStateExceptionData':
        return SetDeviceStateExceptionData(
            settings=[],
            stream_buffers=[],
            pvt_buffers=[],
            triggers=[],
            servo_tuning="",
            stored_positions=[],
            storage=[],
            peripherals=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetDeviceStateExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetDeviceStateExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'settings': [str(item or '') for item in self.settings] if self.settings is not None else [],
            'streamBuffers': [str(item or '') for item in self.stream_buffers] if self.stream_buffers is not None else [],
            'pvtBuffers': [str(item or '') for item in self.pvt_buffers] if self.pvt_buffers is not None else [],
            'triggers': [str(item or '') for item in self.triggers] if self.triggers is not None else [],
            'servoTuning': str(self.servo_tuning or ''),
            'storedPositions': [str(item or '') for item in self.stored_positions] if self.stored_positions is not None else [],
            'storage': [str(item or '') for item in self.storage] if self.storage is not None else [],
            'peripherals': [item.to_dict() for item in self.peripherals] if self.peripherals is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetDeviceStateExceptionData':
        return SetDeviceStateExceptionData(
            settings=data.get('settings'),  # type: ignore
            stream_buffers=data.get('streamBuffers'),  # type: ignore
            pvt_buffers=data.get('pvtBuffers'),  # type: ignore
            triggers=data.get('triggers'),  # type: ignore
            servo_tuning=data.get('servoTuning'),  # type: ignore
            stored_positions=data.get('storedPositions'),  # type: ignore
            storage=data.get('storage'),  # type: ignore
            peripherals=[SetPeripheralStateExceptionData.from_dict(item) for item in data.get('peripherals')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.settings is not None:
            if not isinstance(self.settings, Iterable):
                raise ValueError('Property "Settings" of "SetDeviceStateExceptionData" is not iterable.')

            for i, settings_item in enumerate(self.settings):
                if settings_item is not None:
                    if not isinstance(settings_item, str):
                        raise ValueError(f'Item {i} in property "Settings" of "SetDeviceStateExceptionData" is not a string.')

        if self.stream_buffers is not None:
            if not isinstance(self.stream_buffers, Iterable):
                raise ValueError('Property "StreamBuffers" of "SetDeviceStateExceptionData" is not iterable.')

            for i, stream_buffers_item in enumerate(self.stream_buffers):
                if stream_buffers_item is not None:
                    if not isinstance(stream_buffers_item, str):
                        raise ValueError(f'Item {i} in property "StreamBuffers" of "SetDeviceStateExceptionData" is not a string.')

        if self.pvt_buffers is not None:
            if not isinstance(self.pvt_buffers, Iterable):
                raise ValueError('Property "PvtBuffers" of "SetDeviceStateExceptionData" is not iterable.')

            for i, pvt_buffers_item in enumerate(self.pvt_buffers):
                if pvt_buffers_item is not None:
                    if not isinstance(pvt_buffers_item, str):
                        raise ValueError(f'Item {i} in property "PvtBuffers" of "SetDeviceStateExceptionData" is not a string.')

        if self.triggers is not None:
            if not isinstance(self.triggers, Iterable):
                raise ValueError('Property "Triggers" of "SetDeviceStateExceptionData" is not iterable.')

            for i, triggers_item in enumerate(self.triggers):
                if triggers_item is not None:
                    if not isinstance(triggers_item, str):
                        raise ValueError(f'Item {i} in property "Triggers" of "SetDeviceStateExceptionData" is not a string.')

        if self.servo_tuning is not None:
            if not isinstance(self.servo_tuning, str):
                raise ValueError(f'Property "ServoTuning" of "SetDeviceStateExceptionData" is not a string.')

        if self.stored_positions is not None:
            if not isinstance(self.stored_positions, Iterable):
                raise ValueError('Property "StoredPositions" of "SetDeviceStateExceptionData" is not iterable.')

            for i, stored_positions_item in enumerate(self.stored_positions):
                if stored_positions_item is not None:
                    if not isinstance(stored_positions_item, str):
                        raise ValueError(f'Item {i} in property "StoredPositions" of "SetDeviceStateExceptionData" is not a string.')

        if self.storage is not None:
            if not isinstance(self.storage, Iterable):
                raise ValueError('Property "Storage" of "SetDeviceStateExceptionData" is not iterable.')

            for i, storage_item in enumerate(self.storage):
                if storage_item is not None:
                    if not isinstance(storage_item, str):
                        raise ValueError(f'Item {i} in property "Storage" of "SetDeviceStateExceptionData" is not a string.')

        if self.peripherals is not None:
            if not isinstance(self.peripherals, Iterable):
                raise ValueError('Property "Peripherals" of "SetDeviceStateExceptionData" is not iterable.')

            for i, peripherals_item in enumerate(self.peripherals):
                if peripherals_item is None:
                    raise ValueError(f'Item {i} in property "Peripherals" of "SetDeviceStateExceptionData" is None.')

                if not isinstance(peripherals_item, SetPeripheralStateExceptionData):
                    raise ValueError(f'Item {i} in property "Peripherals" of "SetDeviceStateExceptionData" is not an instance of "SetPeripheralStateExceptionData".')

                peripherals_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class AxisMapping:
    """
    Maps a translator axis to a Zaber stream axis.
    """

    axis_letter: str
    """
    Letter of the translator axis (X,Y,Z,A,B,C,E).
    """

    axis_index: int
    """
    Index of the stream axis.
    """

    @staticmethod
    def zero_values() -> 'AxisMapping':
        return AxisMapping(
            axis_letter="",
            axis_index=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AxisMapping':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AxisMapping.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axisLetter': str(self.axis_letter or ''),
            'axisIndex': int(self.axis_index),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AxisMapping':
        return AxisMapping(
            axis_letter=data.get('axisLetter'),  # type: ignore
            axis_index=data.get('axisIndex'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_letter is not None:
            if not isinstance(self.axis_letter, str):
                raise ValueError(f'Property "AxisLetter" of "AxisMapping" is not a string.')

        if self.axis_index is None:
            raise ValueError(f'Property "AxisIndex" of "AxisMapping" is None.')

        if not isinstance(self.axis_index, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisIndex" of "AxisMapping" is not a number.')

        if int(self.axis_index) != self.axis_index:
            raise ValueError(f'Property "AxisIndex" of "AxisMapping" is not integer value.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

# pylint: disable=too-many-arguments

from typing import TYPE_CHECKING
from ..call import call, call_async, call_sync

from ..dto import requests as dto
from ..dto.binary.binary_settings import BinarySettings
from ..dto.binary.device_identity import DeviceIdentity
from .device_settings import DeviceSettings
from ..dto.binary.command_code import CommandCode
from ..dto.binary.message import Message
from ..units import UnitsAndLiterals, Units, LengthUnits, VelocityUnits
from ..dto.binary.device_type import DeviceType
from ..dto.firmware_version import FirmwareVersion

if TYPE_CHECKING:
    from .connection import Connection


class Device:
    """
    Represents a device using the binary protocol.
    """

    DEFAULT_MOVEMENT_TIMEOUT = 60
    """
    Default timeout for move commands in seconds.
    """

    @property
    def connection(self) -> 'Connection':
        """
        Connection of this device.
        """
        return self._connection

    @property
    def settings(self) -> DeviceSettings:
        """
        Settings and properties of this axis.
        """
        return self._settings

    @property
    def device_address(self) -> int:
        """
        The device address uniquely identifies the device on the connection.
        It can be configured or automatically assigned by the renumber command.
        """
        return self._device_address

    @property
    def identity(self) -> DeviceIdentity:
        """
        Identity of the device.
        """
        return self.__retrieve_identity()

    @property
    def is_identified(self) -> bool:
        """
        Indicates whether or not the device has been identified.
        """
        return self.__retrieve_is_identified()

    @property
    def device_id(self) -> int:
        """
        Unique ID of the device hardware.
        """
        return self.identity.device_id

    @property
    def serial_number(self) -> int:
        """
        Serial number of the device.
        Requires at least Firmware 6.15 for devices or 6.24 for peripherals.
        """
        return self.identity.serial_number

    @property
    def name(self) -> str:
        """
        Name of the product.
        """
        return self.identity.name

    @property
    def firmware_version(self) -> FirmwareVersion:
        """
        Version of the firmware.
        """
        return self.identity.firmware_version

    @property
    def is_peripheral(self) -> bool:
        """
        Indicates whether the device is a peripheral or part of an integrated device.
        """
        return self.identity.is_peripheral

    @property
    def peripheral_id(self) -> int:
        """
        Unique ID of the peripheral hardware.
        """
        return self.identity.peripheral_id

    @property
    def peripheral_name(self) -> str:
        """
        Name of the peripheral hardware.
        """
        return self.identity.peripheral_name

    @property
    def device_type(self) -> DeviceType:
        """
        Determines the type of an device and units it accepts.
        """
        return self.identity.device_type

    def __init__(self, connection: 'Connection', device_address: int):
        self._connection = connection
        self._settings = DeviceSettings(self)
        self._device_address = device_address

    def generic_command(
            self,
            command: CommandCode,
            data: int = 0,
            timeout: float = 0.0,
            check_errors: bool = True
    ) -> Message:
        """
        Sends a generic Binary command to this device.
        For more information please refer to the
        [Binary Protocol Manual](https://www.zaber.com/protocol-manual?protocol=Binary#topic_quick_command_reference).

        Args:
            command: Command to send.
            data: Optional data argument to the command. Defaults to zero.
            timeout: Number of seconds to wait for a response from the device. 0 or negative defaults to 0.5s.
            check_errors: Controls whether to throw an exception when the device rejects the command.

        Returns:
            A response to the command.
        """
        request = dto.GenericBinaryRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
            timeout=timeout,
            check_errors=check_errors,
        )
        response = call(
            "binary/interface/generic_command",
            request,
            Message.from_binary)
        return response

    async def generic_command_async(
            self,
            command: CommandCode,
            data: int = 0,
            timeout: float = 0.0,
            check_errors: bool = True
    ) -> Message:
        """
        Sends a generic Binary command to this device.
        For more information please refer to the
        [Binary Protocol Manual](https://www.zaber.com/protocol-manual?protocol=Binary#topic_quick_command_reference).

        Args:
            command: Command to send.
            data: Optional data argument to the command. Defaults to zero.
            timeout: Number of seconds to wait for a response from the device. 0 or negative defaults to 0.5s.
            check_errors: Controls whether to throw an exception when the device rejects the command.

        Returns:
            A response to the command.
        """
        request = dto.GenericBinaryRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
            timeout=timeout,
            check_errors=check_errors,
        )
        response = await call_async(
            "binary/interface/generic_command",
            request,
            Message.from_binary)
        return response

    def generic_command_no_response(
            self,
            command: CommandCode,
            data: int = 0
    ) -> None:
        """
        Sends a generic Binary command to this device without expecting a response.
        For more information please refer to the
        [Binary Protocol Manual](https://www.zaber.com/protocol-manual?protocol=Binary#topic_quick_command_reference).

        Args:
            command: Command to send.
            data: Optional data argument to the command. Defaults to zero.
        """
        request = dto.GenericBinaryRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
        )
        call("binary/interface/generic_command_no_response", request)

    async def generic_command_no_response_async(
            self,
            command: CommandCode,
            data: int = 0
    ) -> None:
        """
        Sends a generic Binary command to this device without expecting a response.
        For more information please refer to the
        [Binary Protocol Manual](https://www.zaber.com/protocol-manual?protocol=Binary#topic_quick_command_reference).

        Args:
            command: Command to send.
            data: Optional data argument to the command. Defaults to zero.
        """
        request = dto.GenericBinaryRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
        )
        await call_async("binary/interface/generic_command_no_response", request)

    def generic_command_with_units(
            self,
            command: CommandCode,
            data: float = 0,
            from_unit: UnitsAndLiterals = Units.NATIVE,
            to_unit: UnitsAndLiterals = Units.NATIVE,
            timeout: float = 0.0
    ) -> float:
        """
        Sends a generic Binary command to this device with unit conversions for both sent data and retrieved data.

        Args:
            command: Command to send.
            data: Data argument to the command. Defaults to zero.
            from_unit: Unit to convert sent data from.
            to_unit: Unit to convert retrieved data to.
            timeout: Number of seconds to wait for a response from the device. 0 or negative defaults to 0.5s.

        Returns:
            Data that has been converted to the provided unit.
        """
        request = dto.BinaryGenericWithUnitsRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
            from_unit=from_unit,
            to_unit=to_unit,
            timeout=timeout,
        )
        response = call(
            "binary/device/generic_command_with_units",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def generic_command_with_units_async(
            self,
            command: CommandCode,
            data: float = 0,
            from_unit: UnitsAndLiterals = Units.NATIVE,
            to_unit: UnitsAndLiterals = Units.NATIVE,
            timeout: float = 0.0
    ) -> float:
        """
        Sends a generic Binary command to this device with unit conversions for both sent data and retrieved data.

        Args:
            command: Command to send.
            data: Data argument to the command. Defaults to zero.
            from_unit: Unit to convert sent data from.
            to_unit: Unit to convert retrieved data to.
            timeout: Number of seconds to wait for a response from the device. 0 or negative defaults to 0.5s.

        Returns:
            Data that has been converted to the provided unit.
        """
        request = dto.BinaryGenericWithUnitsRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            command=command,
            data=data,
            from_unit=from_unit,
            to_unit=to_unit,
            timeout=timeout,
        )
        response = await call_async(
            "binary/device/generic_command_with_units",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def home(
            self,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Homes device. Device returns to its homing position.

        Args:
            unit: Unit to convert returned position to.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceHomeRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            unit=unit,
            timeout=timeout,
        )
        response = call(
            "binary/device/home",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def home_async(
            self,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Homes device. Device returns to its homing position.

        Args:
            unit: Unit to convert returned position to.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceHomeRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            unit=unit,
            timeout=timeout,
        )
        response = await call_async(
            "binary/device/home",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def stop(
            self,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Stops ongoing device movement. Decelerates until zero speed.

        Args:
            unit: Unit to convert returned position to.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceStopRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            unit=unit,
            timeout=timeout,
        )
        response = call(
            "binary/device/stop",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def stop_async(
            self,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Stops ongoing device movement. Decelerates until zero speed.

        Args:
            unit: Unit to convert returned position to.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceStopRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            unit=unit,
            timeout=timeout,
        )
        response = await call_async(
            "binary/device/stop",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def move_absolute(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Move device to absolute position.

        Args:
            position: Absolute position.
            unit: Unit for the provided position as well as position returned by the device.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.ABS,
            arg=position,
            unit=unit,
            timeout=timeout,
        )
        response = call(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def move_absolute_async(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Move device to absolute position.

        Args:
            position: Absolute position.
            unit: Unit for the provided position as well as position returned by the device.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.ABS,
            arg=position,
            unit=unit,
            timeout=timeout,
        )
        response = await call_async(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def move_relative(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Move device to position relative to current position.

        Args:
            position: Relative position.
            unit: Unit for the provided position as well as position returned by the device.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.REL,
            arg=position,
            unit=unit,
            timeout=timeout,
        )
        response = call(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def move_relative_async(
            self,
            position: float,
            unit: LengthUnits = Units.NATIVE,
            timeout: float = DEFAULT_MOVEMENT_TIMEOUT
    ) -> float:
        """
        Move device to position relative to current position.

        Args:
            position: Relative position.
            unit: Unit for the provided position as well as position returned by the device.
            timeout: Number of seconds to wait for response from the device chain (defaults to 60s).

        Returns:
            Current position that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.REL,
            arg=position,
            unit=unit,
            timeout=timeout,
        )
        response = await call_async(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def move_velocity(
            self,
            velocity: float,
            unit: VelocityUnits = Units.NATIVE
    ) -> float:
        """
        Begins to move device at specified speed.

        Args:
            velocity: Movement velocity.
            unit: Unit to convert returned velocity to.

        Returns:
            Device velocity that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.VEL,
            arg=velocity,
            unit=unit,
        )
        response = call(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def move_velocity_async(
            self,
            velocity: float,
            unit: VelocityUnits = Units.NATIVE
    ) -> float:
        """
        Begins to move device at specified speed.

        Args:
            velocity: Movement velocity.
            unit: Unit to convert returned velocity to.

        Returns:
            Device velocity that has been converted to the provided unit.
        """
        request = dto.BinaryDeviceMoveRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            type=dto.AxisMoveType.VEL,
            arg=velocity,
            unit=unit,
        )
        response = await call_async(
            "binary/device/move",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def wait_until_idle(
            self
    ) -> None:
        """
        Waits until device stops moving.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        call("binary/device/wait_until_idle", request)

    async def wait_until_idle_async(
            self
    ) -> None:
        """
        Waits until device stops moving.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        await call_async("binary/device/wait_until_idle", request)

    def is_busy(
            self
    ) -> bool:
        """
        Check whether the device is moving.

        Returns:
            True if the device is currently executing a motion command.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call(
            "binary/device/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_busy_async(
            self
    ) -> bool:
        """
        Check whether the device is moving.

        Returns:
            True if the device is currently executing a motion command.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = await call_async(
            "binary/device/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def identify(
            self
    ) -> DeviceIdentity:
        """
        Queries the device and the database, gathering information about the product.
        Without this information features such as unit conversions will not work.
        Usually, called automatically by detect devices method.

        Returns:
            Device identification data.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call(
            "binary/device/identify",
            request,
            DeviceIdentity.from_binary)
        return response

    async def identify_async(
            self
    ) -> DeviceIdentity:
        """
        Queries the device and the database, gathering information about the product.
        Without this information features such as unit conversions will not work.
        Usually, called automatically by detect devices method.

        Returns:
            Device identification data.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = await call_async(
            "binary/device/identify",
            request,
            DeviceIdentity.from_binary)
        return response

    def park(
            self
    ) -> None:
        """
        Parks the axis.
        Motor drivers remain enabled and hold current continues to be applied until the device is powered off.
        It can later be unparked and moved without first having to home it.
        Requires at least Firmware 6.06.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        call("binary/device/park", request)

    async def park_async(
            self
    ) -> None:
        """
        Parks the axis.
        Motor drivers remain enabled and hold current continues to be applied until the device is powered off.
        It can later be unparked and moved without first having to home it.
        Requires at least Firmware 6.06.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        await call_async("binary/device/park", request)

    def unpark(
            self
    ) -> None:
        """
        Unparks axis. Axis will now be able to move.
        Requires at least Firmware 6.06.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        call("binary/device/unpark", request)

    async def unpark_async(
            self
    ) -> None:
        """
        Unparks axis. Axis will now be able to move.
        Requires at least Firmware 6.06.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        await call_async("binary/device/unpark", request)

    def is_parked(
            self
    ) -> bool:
        """
        Returns bool indicating whether the axis is parked or not.
        Requires at least Firmware 6.06.

        Returns:
            True if the axis is currently parked. False otherwise.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call(
            "binary/device/is_parked",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_parked_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether the axis is parked or not.
        Requires at least Firmware 6.06.

        Returns:
            True if the axis is currently parked. False otherwise.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = await call_async(
            "binary/device/is_parked",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def get_position(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Returns current device position.

        Args:
            unit: Units of position.

        Returns:
            Axis position.
        """
        request = dto.BinaryDeviceGetSettingRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            setting=BinarySettings.CURRENT_POSITION,
            unit=unit,
        )
        response = call(
            "binary/device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_position_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Returns current device position.

        Args:
            unit: Units of position.

        Returns:
            Axis position.
        """
        request = dto.BinaryDeviceGetSettingRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
            setting=BinarySettings.CURRENT_POSITION,
            unit=unit,
        )
        response = await call_async(
            "binary/device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call_sync(
            "binary/device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def __retrieve_identity(
            self
    ) -> DeviceIdentity:
        """
        Returns identity.

        Returns:
            Device identity.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call_sync(
            "binary/device/get_identity",
            request,
            DeviceIdentity.from_binary)
        return response

    def __retrieve_is_identified(
            self
    ) -> bool:
        """
        Returns whether or not the device have been identified.

        Returns:
            True if the device has already been identified. False otherwise.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.connection.interface_id,
            device=self.device_address,
        )
        response = call_sync(
            "binary/device/get_is_identified",
            request,
            dto.BoolResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .binary_command_failed_exception_data import BinaryCommandFailedExceptionData as BinaryCommandFailedExceptionData
from .command_failed_exception_data import CommandFailedExceptionData as CommandFailedExceptionData
from .command_too_long_exception_data import CommandTooLongExceptionData as CommandTooLongExceptionData
from .device_address_conflict_exception_data import DeviceAddressConflictExceptionData as DeviceAddressConflictExceptionData
from .device_db_failed_exception_data import DeviceDbFailedExceptionData as DeviceDbFailedExceptionData
from .g_code_execution_exception_data import GCodeExecutionExceptionData as GCodeExecutionExceptionData
from .g_code_syntax_exception_data import GCodeSyntaxExceptionData as GCodeSyntaxExceptionData
from .invalid_packet_exception_data import InvalidPacketExceptionData as InvalidPacketExceptionData
from .invalid_pvt_point import InvalidPvtPoint as InvalidPvtPoint
from .invalid_response_exception_data import InvalidResponseExceptionData as InvalidResponseExceptionData
from .movement_failed_exception_data import MovementFailedExceptionData as MovementFailedExceptionData
from .movement_interrupted_exception_data import MovementInterruptedExceptionData as MovementInterruptedExceptionData
from .operation_failed_exception_data import OperationFailedExceptionData as OperationFailedExceptionData
from .pvt_execution_exception_data import PvtExecutionExceptionData as PvtExecutionExceptionData
from .pvt_movement_failed_exception_data import PvtMovementFailedExceptionData as PvtMovementFailedExceptionData
from .pvt_movement_interrupted_exception_data import PvtMovementInterruptedExceptionData as PvtMovementInterruptedExceptionData
from .set_device_state_exception_data import SetDeviceStateExceptionData as SetDeviceStateExceptionData
from .set_peripheral_state_exception_data import SetPeripheralStateExceptionData as SetPeripheralStateExceptionData
from .stream_execution_exception_data import StreamExecutionExceptionData as StreamExecutionExceptionData
from .stream_movement_failed_exception_data import StreamMovementFailedExceptionData as StreamMovementFailedExceptionData
from .stream_movement_interrupted_exception_data import StreamMovementInterruptedExceptionData as StreamMovementInterruptedExceptionData



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Union
from ..dto.exceptions.invalid_packet_exception_data import InvalidPacketExceptionData
from .motion_lib_exception import MotionLibException


class InvalidPacketException(MotionLibException):
    """
    Thrown when a packet from a device cannot be parsed.
    """

    @property
    def details(self) -> InvalidPacketExceptionData:
        """
        Additional data for InvalidPacketException
        """
        return self._details

    def __init__(self, message: str, custom_data: Union[bytes, InvalidPacketExceptionData]):
        MotionLibException.__init__(self, message)

        if isinstance(custom_data, InvalidPacketExceptionData):
            self._details = custom_data
        else:
            self._details = InvalidPacketExceptionData.from_binary(custom_data)



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Union
from ..dto.exceptions.operation_failed_exception_data import OperationFailedExceptionData
from .motion_lib_exception import MotionLibException


class OperationFailedException(MotionLibException):
    """
    Thrown when a non-motion device fails to perform a requested operation.
    """

    @property
    def details(self) -> OperationFailedExceptionData:
        """
        Additional data for OperationFailedException
        """
        return self._details

    def __init__(self, message: str, custom_data: Union[bytes, OperationFailedExceptionData]):
        MotionLibException.__init__(self, message)

        if isinstance(custom_data, OperationFailedExceptionData):
            self._details = custom_data
        else:
            self._details = OperationFailedExceptionData.from_binary(custom_data)



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson
from ..ascii.get_setting_result import GetSettingResult


@dataclass
class GetSettingResults:

    results: List[GetSettingResult] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'GetSettingResults':
        return GetSettingResults(
            results=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GetSettingResults':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GetSettingResults.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'results': [item.to_dict() for item in self.results] if self.results is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GetSettingResults':
        return GetSettingResults(
            results=[GetSettingResult.from_dict(item) for item in data.get('results')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.results is not None:
            if not isinstance(self.results, Iterable):
                raise ValueError('Property "Results" of "GetSettingResults" is not iterable.')

            for i, results_item in enumerate(self.results):
                if results_item is None:
                    raise ValueError(f'Item {i} in property "Results" of "GetSettingResults" is None.')

                if not isinstance(results_item, GetSettingResult):
                    raise ValueError(f'Item {i} in property "Results" of "GetSettingResults" is not an instance of "GetSettingResult".')

                results_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import zaber_bson
from ..ascii.stream_mode import StreamMode
from ..ascii.pvt_mode import PvtMode


@dataclass
class StreamModeResponse:

    stream_mode: StreamMode = next(first for first in StreamMode)

    pvt_mode: PvtMode = next(first for first in PvtMode)

    @staticmethod
    def zero_values() -> 'StreamModeResponse':
        return StreamModeResponse(
            stream_mode=next(first for first in StreamMode),
            pvt_mode=next(first for first in PvtMode),
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'StreamModeResponse':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return StreamModeResponse.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'streamMode': self.stream_mode.value,
            'pvtMode': self.pvt_mode.value,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'StreamModeResponse':
        return StreamModeResponse(
            stream_mode=StreamMode(data.get('streamMode')),  # type: ignore
            pvt_mode=PvtMode(data.get('pvtMode')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.stream_mode is None:
            raise ValueError(f'Property "StreamMode" of "StreamModeResponse" is None.')

        if not isinstance(self.stream_mode, StreamMode):
            raise ValueError(f'Property "StreamMode" of "StreamModeResponse" is not an instance of "StreamMode".')

        if self.pvt_mode is None:
            raise ValueError(f'Property "PvtMode" of "StreamModeResponse" is None.')

        if not isinstance(self.pvt_mode, PvtMode):
            raise ValueError(f'Property "PvtMode" of "StreamModeResponse" is not an instance of "PvtMode".')



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .binary_command_failed_exception import BinaryCommandFailedException as BinaryCommandFailedException
from .command_failed_exception import CommandFailedException as CommandFailedException
from .command_preempted_exception import CommandPreemptedException as CommandPreemptedException
from .command_too_long_exception import CommandTooLongException as CommandTooLongException
from .connection_closed_exception import ConnectionClosedException as ConnectionClosedException
from .connection_failed_exception import ConnectionFailedException as ConnectionFailedException
from .conversion_failed_exception import ConversionFailedException as ConversionFailedException
from .device_address_conflict_exception import DeviceAddressConflictException as DeviceAddressConflictException
from .device_busy_exception import DeviceBusyException as DeviceBusyException
from .device_db_failed_exception import DeviceDbFailedException as DeviceDbFailedException
from .device_detection_failed_exception import DeviceDetectionFailedException as DeviceDetectionFailedException
from .device_failed_exception import DeviceFailedException as DeviceFailedException
from .device_not_identified_exception import DeviceNotIdentifiedException as DeviceNotIdentifiedException
from .g_code_execution_exception import GCodeExecutionException as GCodeExecutionException
from .g_code_syntax_exception import GCodeSyntaxException as GCodeSyntaxException
from .incompatible_shared_library_exception import IncompatibleSharedLibraryException as IncompatibleSharedLibraryException
from .internal_error_exception import InternalErrorException as InternalErrorException
from .invalid_argument_exception import InvalidArgumentException as InvalidArgumentException
from .invalid_data_exception import InvalidDataException as InvalidDataException
from .invalid_operation_exception import InvalidOperationException as InvalidOperationException
from .invalid_packet_exception import InvalidPacketException as InvalidPacketException
from .invalid_park_state_exception import InvalidParkStateException as InvalidParkStateException
from .invalid_request_data_exception import InvalidRequestDataException as InvalidRequestDataException
from .invalid_response_exception import InvalidResponseException as InvalidResponseException
from .io_channel_out_of_range_exception import IoChannelOutOfRangeException as IoChannelOutOfRangeException
from .io_failed_exception import IoFailedException as IoFailedException
from .lockstep_enabled_exception import LockstepEnabledException as LockstepEnabledException
from .lockstep_not_enabled_exception import LockstepNotEnabledException as LockstepNotEnabledException
from .motion_lib_exception import MotionLibException as MotionLibException
from .movement_failed_exception import MovementFailedException as MovementFailedException
from .movement_interrupted_exception import MovementInterruptedException as MovementInterruptedException
from .no_device_found_exception import NoDeviceFoundException as NoDeviceFoundException
from .no_value_for_key_exception import NoValueForKeyException as NoValueForKeyException
from .not_supported_exception import NotSupportedException as NotSupportedException
from .operation_failed_exception import OperationFailedException as OperationFailedException
from .os_failed_exception import OsFailedException as OsFailedException
from .out_of_request_ids_exception import OutOfRequestIdsException as OutOfRequestIdsException
from .pvt_discontinuity_exception import PvtDiscontinuityException as PvtDiscontinuityException
from .pvt_execution_exception import PvtExecutionException as PvtExecutionException
from .pvt_mode_exception import PvtModeException as PvtModeException
from .pvt_movement_failed_exception import PvtMovementFailedException as PvtMovementFailedException
from .pvt_movement_interrupted_exception import PvtMovementInterruptedException as PvtMovementInterruptedException
from .pvt_setup_failed_exception import PvtSetupFailedException as PvtSetupFailedException
from .request_timeout_exception import RequestTimeoutException as RequestTimeoutException
from .serial_port_busy_exception import SerialPortBusyException as SerialPortBusyException
from .set_device_state_failed_exception import SetDeviceStateFailedException as SetDeviceStateFailedException
from .set_peripheral_state_failed_exception import SetPeripheralStateFailedException as SetPeripheralStateFailedException
from .setting_not_found_exception import SettingNotFoundException as SettingNotFoundException
from .stream_discontinuity_exception import StreamDiscontinuityException as StreamDiscontinuityException
from .stream_execution_exception import StreamExecutionException as StreamExecutionException
from .stream_mode_exception import StreamModeException as StreamModeException
from .stream_movement_failed_exception import StreamMovementFailedException as StreamMovementFailedException
from .stream_movement_interrupted_exception import StreamMovementInterruptedException as StreamMovementInterruptedException
from .stream_setup_failed_exception import StreamSetupFailedException as StreamSetupFailedException
from .timeout_exception import TimeoutException as TimeoutException
from .transport_already_used_exception import TransportAlreadyUsedException as TransportAlreadyUsedException
from .unknown_request_exception import UnknownRequestException as UnknownRequestException
from ..dto.exceptions.binary_command_failed_exception_data import BinaryCommandFailedExceptionData as BinaryCommandFailedExceptionData
from ..dto.exceptions.command_failed_exception_data import CommandFailedExceptionData as CommandFailedExceptionData
from ..dto.exceptions.command_too_long_exception_data import CommandTooLongExceptionData as CommandTooLongExceptionData
from ..dto.exceptions.device_address_conflict_exception_data import DeviceAddressConflictExceptionData as DeviceAddressConflictExceptionData
from ..dto.exceptions.device_db_failed_exception_data import DeviceDbFailedExceptionData as DeviceDbFailedExceptionData
from ..dto.exceptions.g_code_execution_exception_data import GCodeExecutionExceptionData as GCodeExecutionExceptionData
from ..dto.exceptions.g_code_syntax_exception_data import GCodeSyntaxExceptionData as GCodeSyntaxExceptionData
from ..dto.exceptions.invalid_packet_exception_data import InvalidPacketExceptionData as InvalidPacketExceptionData
from ..dto.exceptions.invalid_pvt_point import InvalidPvtPoint as InvalidPvtPoint
from ..dto.exceptions.invalid_response_exception_data import InvalidResponseExceptionData as InvalidResponseExceptionData
from ..dto.exceptions.movement_failed_exception_data import MovementFailedExceptionData as MovementFailedExceptionData
from ..dto.exceptions.movement_interrupted_exception_data import MovementInterruptedExceptionData as MovementInterruptedExceptionData
from ..dto.exceptions.operation_failed_exception_data import OperationFailedExceptionData as OperationFailedExceptionData
from ..dto.exceptions.pvt_execution_exception_data import PvtExecutionExceptionData as PvtExecutionExceptionData
from ..dto.exceptions.pvt_movement_failed_exception_data import PvtMovementFailedExceptionData as PvtMovementFailedExceptionData
from ..dto.exceptions.pvt_movement_interrupted_exception_data import PvtMovementInterruptedExceptionData as PvtMovementInterruptedExceptionData
from ..dto.exceptions.set_device_state_exception_data import SetDeviceStateExceptionData as SetDeviceStateExceptionData
from ..dto.exceptions.set_peripheral_state_exception_data import SetPeripheralStateExceptionData as SetPeripheralStateExceptionData
from ..dto.exceptions.stream_execution_exception_data import StreamExecutionExceptionData as StreamExecutionExceptionData
from ..dto.exceptions.stream_movement_failed_exception_data import StreamMovementFailedExceptionData as StreamMovementFailedExceptionData
from ..dto.exceptions.stream_movement_interrupted_exception_data import StreamMovementInterruptedExceptionData as StreamMovementInterruptedExceptionData



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List

from ..call import call, call_async
from ..dto import requests as dto

from .pvt_sequence import PvtSequence
from .pvt_buffer import PvtBuffer

if TYPE_CHECKING:
    from .device import Device


class Pvt:
    """
    Class providing access to device PVT (Position-Velocity-Time) features.
    Requires at least Firmware 7.33.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that this PVT belongs to.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def get_sequence(
            self,
            pvt_id: int
    ) -> 'PvtSequence':
        """
        Gets a PvtSequence class instance which allows you to control a particular PVT sequence on the device.

        Args:
            pvt_id: The ID of the PVT sequence to control. The IDs start at 1.

        Returns:
            PvtSequence instance.
        """
        if pvt_id <= 0:
            raise ValueError('Invalid value; PVT sequences are numbered from 1.')

        return PvtSequence(self.device, pvt_id)

    def get_buffer(
            self,
            pvt_buffer_id: int
    ) -> 'PvtBuffer':
        """
        Gets a PvtBuffer class instance which is a handle for a PVT buffer on the device.

        Args:
            pvt_buffer_id: The ID of the PVT buffer to control. PVT buffer IDs start at one.

        Returns:
            PvtBuffer instance.
        """
        if pvt_buffer_id <= 0:
            raise ValueError('Invalid value; PVT buffers are numbered from 1.')

        return PvtBuffer(self.device, pvt_buffer_id)

    def list_buffer_ids(
            self
    ) -> List[int]:
        """
        Get a list of buffer IDs that are currently in use.

        Returns:
            List of buffer IDs.
        """
        request = dto.StreamBufferList(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            pvt=True,
        )
        response = call(
            "device/stream_buffer_list",
            request,
            dto.IntArrayResponse.from_binary)
        return response.values

    async def list_buffer_ids_async(
            self
    ) -> List[int]:
        """
        Get a list of buffer IDs that are currently in use.

        Returns:
            List of buffer IDs.
        """
        request = dto.StreamBufferList(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            pvt=True,
        )
        response = await call_async(
            "device/stream_buffer_list",
            request,
            dto.IntArrayResponse.from_binary)
        return response.values



# This file is generated. Do not modify by hand.
from enum import Enum


class Errors(Enum):

    REQUEST_TIMEOUT = 0
    CONNECTION_CLOSED = 1
    INVALID_ARGUMENT = 2
    OUT_OF_REQUEST_IDS = 3
    INVALID_PACKET = 4
    CONNECTION_FAILED = 5
    UNKNOWN_REQUEST = 6
    INVALID_REQUEST_DATA = 7
    COMMAND_FAILED = 8
    DEVICE_DB_FAILED = 9
    INVALID_DATA = 10
    DEVICE_NOT_IDENTIFIED = 11
    CONVERSION_FAILED = 12
    DEVICE_ADDRESS_CONFLICT = 13
    NO_DEVICE_FOUND = 14
    MOVEMENT_INTERRUPTED = 15
    MOVEMENT_FAILED = 16
    IO_FAILED = 17
    INVALID_RESPONSE = 18
    NOT_SUPPORTED = 19
    DEVICE_FAILED = 20
    OS_FAILED = 21
    INTERNAL_ERROR = 22
    BINARY_COMMAND_FAILED = 23
    COMMAND_PREEMPTED = 24
    LOCKSTEP_NOT_ENABLED = 25
    LOCKSTEP_ENABLED = 26
    IO_CHANNEL_OUT_OF_RANGE = 27
    SETTING_NOT_FOUND = 28
    STREAM_MODE = 29
    STREAM_EXECUTION = 30
    STREAM_MOVEMENT_FAILED = 31
    STREAM_SETUP_FAILED = 32
    DEVICE_BUSY = 33
    STREAM_MOVEMENT_INTERRUPTED = 34
    INVALID_PARK_STATE = 35
    SERIAL_PORT_BUSY = 36
    TRANSPORT_ALREADY_USED = 37
    SET_PERIPHERAL_STATE_FAILED = 38
    SET_DEVICE_STATE_FAILED = 39
    G_CODE_SYNTAX = 40
    G_CODE_EXECUTION = 41
    INVALID_OPERATION = 42
    COMMAND_TOO_LONG = 43
    NO_VALUE_FOR_KEY = 44
    DEVICE_DETECTION_FAILED = 45
    TIMEOUT = 46
    PVT_MODE = 47
    PVT_EXECUTION = 48
    PVT_SETUP_FAILED = 49
    PVT_MOVEMENT_FAILED = 50
    PVT_MOVEMENT_INTERRUPTED = 51
    PVT_DISCONTINUITY = 52
    STREAM_DISCONTINUITY = 53
    OPERATION_FAILED = 54
    INCOMPATIBLE_SHARED_LIBRARY = 55



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import zaber_bson
from ..device_db_source_type import DeviceDbSourceType


@dataclass
class SetDeviceDbSourceRequest:

    source_type: DeviceDbSourceType = next(first for first in DeviceDbSourceType)

    url_or_file_path: Optional[str] = None

    @staticmethod
    def zero_values() -> 'SetDeviceDbSourceRequest':
        return SetDeviceDbSourceRequest(
            source_type=next(first for first in DeviceDbSourceType),
            url_or_file_path=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetDeviceDbSourceRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetDeviceDbSourceRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'sourceType': self.source_type.value,
            'urlOrFilePath': str(self.url_or_file_path) if self.url_or_file_path is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetDeviceDbSourceRequest':
        return SetDeviceDbSourceRequest(
            source_type=DeviceDbSourceType(data.get('sourceType')),  # type: ignore
            url_or_file_path=data.get('urlOrFilePath'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.source_type is None:
            raise ValueError(f'Property "SourceType" of "SetDeviceDbSourceRequest" is None.')

        if not isinstance(self.source_type, DeviceDbSourceType):
            raise ValueError(f'Property "SourceType" of "SetDeviceDbSourceRequest" is not an instance of "DeviceDbSourceType".')

        if self.url_or_file_path is not None:
            if not isinstance(self.url_or_file_path, str):
                raise ValueError(f'Property "UrlOrFilePath" of "SetDeviceDbSourceRequest" is not a string.')



# This file is generated. Do not modify by hand.
from enum import Enum


class ReplyCode(Enum):
    """
    Named constants for all Zaber Binary protocol reply-only command codes.
    """

    MOVE_TRACKING = 8
    LIMIT_ACTIVE = 9
    MANUAL_MOVE_TRACKING = 10
    MANUAL_MOVE = 11
    SLIP_TRACKING = 12
    UNEXPECTED_POSITION = 13
    ERROR = 255



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .motion_lib_exception import MotionLibException


class IncompatibleSharedLibraryException(MotionLibException):
    """
    Thrown when the loaded shared library is incompatible with the running code.
    Typically caused by mixed library binary files. Reinstall the library.
    """



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import zaber_bson
from .response_type import ResponseType
from .errors import Errors


@dataclass
class GatewayResponse:

    response: ResponseType = next(first for first in ResponseType)

    error_type: Errors = next(first for first in Errors)

    error_message: str = ""

    @staticmethod
    def zero_values() -> 'GatewayResponse':
        return GatewayResponse(
            response=next(first for first in ResponseType),
            error_type=next(first for first in Errors),
            error_message="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GatewayResponse':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GatewayResponse.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'response': self.response.value,
            'errorType': self.error_type.value,
            'errorMessage': str(self.error_message or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GatewayResponse':
        return GatewayResponse(
            response=ResponseType(data.get('response')),  # type: ignore
            error_type=Errors(data.get('errorType')),  # type: ignore
            error_message=data.get('errorMessage'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.response is None:
            raise ValueError(f'Property "Response" of "GatewayResponse" is None.')

        if not isinstance(self.response, ResponseType):
            raise ValueError(f'Property "Response" of "GatewayResponse" is not an instance of "ResponseType".')

        if self.error_type is None:
            raise ValueError(f'Property "ErrorType" of "GatewayResponse" is None.')

        if not isinstance(self.error_type, Errors):
            raise ValueError(f'Property "ErrorType" of "GatewayResponse" is not an instance of "Errors".')

        if self.error_message is not None:
            if not isinstance(self.error_message, str):
                raise ValueError(f'Property "ErrorMessage" of "GatewayResponse" is not a string.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import List
from ..dto import requests as dto
from ..call import call_sync
from ..units import UnitsAndLiterals, Units, TimeUnits, FrequencyUnits
from ..dto.ascii.io_port_type import IoPortType
from ..dto.ascii.oscilloscope_data_source import OscilloscopeDataSource
from ..dto.ascii.oscilloscope_capture_properties import OscilloscopeCaptureProperties


class OscilloscopeData:
    """
    Contains a block of contiguous recorded data for one channel of the device's oscilloscope.
    """

    @property
    def data_id(self) -> int:
        """
        Unique ID for this block of recorded data.
        """
        return self._data_id

    @property
    def data_source(self) -> OscilloscopeDataSource:
        """
        Indicates whether the data came from a setting or an I/O pin.
        """
        return self.__retrieve_properties().data_source

    @property
    def setting(self) -> str:
        """
        The name of the recorded setting.
        """
        return self.__retrieve_properties().setting

    @property
    def axis_number(self) -> int:
        """
        The number of the axis the data was recorded from, or 0 for the controller.
        """
        return self.__retrieve_properties().axis_number

    @property
    def io_type(self) -> IoPortType:
        """
        Which kind of I/O port data was recorded from.
        """
        return self.__retrieve_properties().io_type

    @property
    def io_channel(self) -> int:
        """
        Which I/O pin within the port was recorded.
        """
        return self.__retrieve_properties().io_channel

    def __init__(self, data_id: int):
        self._data_id = data_id

    def get_timebase(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the sample interval that this data was recorded with.

        Args:
            unit: Unit of measure to represent the timebase in.

        Returns:
            The timebase setting at the time the data was recorded.
        """
        request = dto.OscilloscopeDataGetRequest(
            data_id=self.data_id,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_timebase",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_frequency(
            self,
            unit: FrequencyUnits = Units.NATIVE
    ) -> float:
        """
        Get the sampling frequency that this data was recorded with.

        Args:
            unit: Unit of measure to represent the frequency in.

        Returns:
            The frequency (inverse of the timebase setting) at the time the data was recorded.
        """
        request = dto.OscilloscopeDataGetRequest(
            data_id=self.data_id,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_frequency",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_delay(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the user-specified time period between receipt of the start command and the first data point.
        Under some circumstances, the actual delay may be different - call GetSampleTime(0) to get the effective delay.

        Args:
            unit: Unit of measure to represent the delay in.

        Returns:
            The delay setting at the time the data was recorded.
        """
        request = dto.OscilloscopeDataGetRequest(
            data_id=self.data_id,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_delay",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_sample_time(
            self,
            index: int,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Calculate the time a sample was recorded, relative to when the recording was triggered.

        Args:
            index: 0-based index of the sample to calculate the time of.
            unit: Unit of measure to represent the calculated time in.

        Returns:
            The calculated time offset of the data sample at the given index.
        """
        request = dto.OscilloscopeDataGetSampleTimeRequest(
            data_id=self.data_id,
            index=index,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_sample_time",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_sample_times(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> List[float]:
        """
        Calculate the time for all samples, relative to when the recording was triggered.

        Args:
            unit: Unit of measure to represent the calculated time in.

        Returns:
            The calculated time offsets of all data samples.
        """
        request = dto.OscilloscopeDataGetSampleTimeRequest(
            data_id=self.data_id,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_sample_times",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    def get_data(
            self,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> List[float]:
        """
        Get the recorded data as an array of doubles.

        Args:
            unit: Unit of measure to convert the data to.

        Returns:
            The recorded data for one oscilloscope channel, converted to the units specified.
        """
        request = dto.OscilloscopeDataGetRequest(
            data_id=self.data_id,
            unit=unit,
        )
        response = call_sync(
            "oscilloscopedata/get_samples",
            request,
            dto.OscilloscopeDataGetSamplesResponse.from_binary)
        return response.data

    @staticmethod
    def __free(
            data_id: int
    ) -> None:
        """
        Releases native resources of an oscilloscope data buffer.

        Args:
            data_id: The ID of the data buffer to delete.
        """
        request = dto.OscilloscopeDataIdentifier(
            data_id=data_id,
        )
        call_sync("oscilloscopedata/free", request)

    def __retrieve_properties(
            self
    ) -> OscilloscopeCaptureProperties:
        """
        Returns recording properties.

        Returns:
            Capture properties.
        """
        request = dto.OscilloscopeDataIdentifier(
            data_id=self.data_id,
        )
        response = call_sync(
            "oscilloscopedata/get_properties",
            request,
            OscilloscopeCaptureProperties.from_binary)
        return response

    def __del__(self) -> None:
        OscilloscopeData.__free(self.data_id)



# This file is generated. Do not modify by hand.
from enum import Enum


class StreamMode(Enum):
    """
    Mode of a stream.
    """

    DISABLED = 0
    STORE = 1
    STORE_ARBITRARY_AXES = 2
    LIVE = 3



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .connection import Connection as Connection
from .device import Device as Device
from .device_settings import DeviceSettings as DeviceSettings
from ..dto.binary.binary_settings import BinarySettings as BinarySettings
from ..dto.binary.command_code import CommandCode as CommandCode
from ..dto.binary.device_identity import DeviceIdentity as DeviceIdentity
from ..dto.binary.device_type import DeviceType as DeviceType
from ..dto.binary.error_code import ErrorCode as ErrorCode
from ..dto.binary.message import Message as Message
from ..dto.binary.reply_code import ReplyCode as ReplyCode
from ..dto.binary.reply_only_event import ReplyOnlyEvent as ReplyOnlyEvent
from ..dto.binary.unknown_response_event import UnknownResponseEvent as UnknownResponseEvent



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .call import call_sync
from .dto import requests as dto
from .units import UnitsAndLiterals


class UnitTable:
    """
    Helper for working with units of measure.
    """

    @staticmethod
    def get_symbol(
            unit: UnitsAndLiterals
    ) -> str:
        """
        Gets the standard symbol associated with a given unit.

        Args:
            unit: Unit of measure.

        Returns:
            Symbols corresponding to the given unit. Throws NoValueForKey if no symbol is defined.
        """
        request = dto.UnitGetSymbolRequest(
            unit=unit,
        )
        response = call_sync(
            "units/get_symbol",
            request,
            dto.UnitGetSymbolResponse.from_binary)
        return response.symbol

    @staticmethod
    def get_unit(
            symbol: str
    ) -> UnitsAndLiterals:
        """
        Gets the unit enum value associated with a standard symbol.
        Note not all units can be retrieved this way.

        Args:
            symbol: Symbol to look up.

        Returns:
            The unit enum value with the given symbols. Throws NoValueForKey if the symbol is not supported for lookup.
        """
        request = dto.UnitGetEnumRequest(
            symbol=symbol,
        )
        response = call_sync(
            "units/get_enum",
            request,
            dto.UnitGetEnumResponse.from_binary)
        return response.unit



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..dto import requests as dto
from ..units import Units, TimeUnits, FrequencyUnits
from ..call import call, call_async
from ..dto.ascii.io_port_type import IoPortType
from .oscilloscope_data import OscilloscopeData

if TYPE_CHECKING:
    from .device import Device


class Oscilloscope:
    """
    Provides a convenient way to control the oscilloscope data recording feature of some devices.
    The oscilloscope can record the values of some settings over time at high resolution.
    Requires at least Firmware 7.00.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that this Oscilloscope measures.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def add_channel(
            self,
            axis: int,
            setting: str
    ) -> None:
        """
        Select a setting to be recorded.

        Args:
            axis: The 1-based index of the axis to record the value from.
            setting: The name of a setting to record.
        """
        request = dto.OscilloscopeAddSettingChannelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=axis,
            setting=setting,
        )
        call("oscilloscope/add_setting_channel", request)

    async def add_channel_async(
            self,
            axis: int,
            setting: str
    ) -> None:
        """
        Select a setting to be recorded.

        Args:
            axis: The 1-based index of the axis to record the value from.
            setting: The name of a setting to record.
        """
        request = dto.OscilloscopeAddSettingChannelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            axis=axis,
            setting=setting,
        )
        await call_async("oscilloscope/add_setting_channel", request)

    def add_io_channel(
            self,
            io_type: IoPortType,
            io_channel: int
    ) -> None:
        """
        Select an I/O pin to be recorded.
        Requires at least Firmware 7.33.

        Args:
            io_type: The I/O port type to read data from.
            io_channel: The 1-based index of the I/O pin to read from.
        """
        request = dto.OscilloscopeAddIoChannelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            io_type=io_type,
            io_channel=io_channel,
        )
        call("oscilloscope/add_io_channel", request)

    async def add_io_channel_async(
            self,
            io_type: IoPortType,
            io_channel: int
    ) -> None:
        """
        Select an I/O pin to be recorded.
        Requires at least Firmware 7.33.

        Args:
            io_type: The I/O port type to read data from.
            io_channel: The 1-based index of the I/O pin to read from.
        """
        request = dto.OscilloscopeAddIoChannelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            io_type=io_type,
            io_channel=io_channel,
        )
        await call_async("oscilloscope/add_io_channel", request)

    def clear(
            self
    ) -> None:
        """
        Clear the list of channels to record.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        call("oscilloscope/clear_channels", request)

    async def clear_async(
            self
    ) -> None:
        """
        Clear the list of channels to record.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        await call_async("oscilloscope/clear_channels", request)

    def get_timebase(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the current sampling interval.

        Args:
            unit: Unit of measure to represent the timebase in.

        Returns:
            The current sampling interval in the selected time units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            unit=unit,
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_timebase_async(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the current sampling interval.

        Args:
            unit: Unit of measure to represent the timebase in.

        Returns:
            The current sampling interval in the selected time units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            unit=unit,
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_timebase(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling interval.

        Args:
            interval: Sample interval for the next oscilloscope recording. Minimum value is 100µs.
            unit: Unit of measure the timebase is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            value=interval,
            unit=unit,
        )
        call("device/set_setting", request)

    async def set_timebase_async(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling interval.

        Args:
            interval: Sample interval for the next oscilloscope recording. Minimum value is 100µs.
            unit: Unit of measure the timebase is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            value=interval,
            unit=unit,
        )
        await call_async("device/set_setting", request)

    def get_frequency(
            self,
            unit: FrequencyUnits = Units.NATIVE
    ) -> float:
        """
        Get the current sampling frequency.
        The values is calculated as the inverse of the current sampling interval.

        Args:
            unit: Unit of measure to represent the frequency in.

        Returns:
            The inverse of current sampling interval in the selected units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            unit=unit,
        )
        response = call(
            "oscilloscope/get_frequency",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_frequency_async(
            self,
            unit: FrequencyUnits = Units.NATIVE
    ) -> float:
        """
        Get the current sampling frequency.
        The values is calculated as the inverse of the current sampling interval.

        Args:
            unit: Unit of measure to represent the frequency in.

        Returns:
            The inverse of current sampling interval in the selected units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            unit=unit,
        )
        response = await call_async(
            "oscilloscope/get_frequency",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_frequency(
            self,
            frequency: float,
            unit: FrequencyUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling frequency (inverse of the sampling interval).
        The value is quantized to the next closest value supported by the firmware.

        Args:
            frequency: Sample frequency for the next oscilloscope recording.
            unit: Unit of measure the frequency is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            value=frequency,
            unit=unit,
        )
        call("oscilloscope/set_frequency", request)

    async def set_frequency_async(
            self,
            frequency: float,
            unit: FrequencyUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling frequency (inverse of the sampling interval).
        The value is quantized to the next closest value supported by the firmware.

        Args:
            frequency: Sample frequency for the next oscilloscope recording.
            unit: Unit of measure the frequency is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.timebase",
            value=frequency,
            unit=unit,
        )
        await call_async("oscilloscope/set_frequency", request)

    def get_delay(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the delay before oscilloscope recording starts.

        Args:
            unit: Unit of measure to represent the delay in.

        Returns:
            The current start delay in the selected time units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.delay",
            unit=unit,
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_delay_async(
            self,
            unit: TimeUnits = Units.NATIVE
    ) -> float:
        """
        Get the delay before oscilloscope recording starts.

        Args:
            unit: Unit of measure to represent the delay in.

        Returns:
            The current start delay in the selected time units.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.delay",
            unit=unit,
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_delay(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling start delay.

        Args:
            interval: Delay time between triggering a recording and the first data point being recorded.
            unit: Unit of measure the delay is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.delay",
            value=interval,
            unit=unit,
        )
        call("device/set_setting", request)

    async def set_delay_async(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set the sampling start delay.

        Args:
            interval: Delay time between triggering a recording and the first data point being recorded.
            unit: Unit of measure the delay is represented in.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.delay",
            value=interval,
            unit=unit,
        )
        await call_async("device/set_setting", request)

    def get_max_channels(
            self
    ) -> int:
        """
        Get the maximum number of channels that can be recorded.

        Returns:
            The maximum number of channels that can be added to an Oscilloscope recording.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.numchannels",
        )
        response = call(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_max_channels_async(
            self
    ) -> int:
        """
        Get the maximum number of channels that can be recorded.

        Returns:
            The maximum number of channels that can be added to an Oscilloscope recording.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.numchannels",
        )
        response = await call_async(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_max_buffer_size(
            self
    ) -> int:
        """
        Get the maximum number of samples that can be recorded per Oscilloscope channel.

        Returns:
            The maximum number of samples that can be recorded per Oscilloscope channel.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.channel.size.max",
        )
        response = call(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_max_buffer_size_async(
            self
    ) -> int:
        """
        Get the maximum number of samples that can be recorded per Oscilloscope channel.

        Returns:
            The maximum number of samples that can be recorded per Oscilloscope channel.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.channel.size.max",
        )
        response = await call_async(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_buffer_size(
            self
    ) -> int:
        """
        Get the number of samples that can be recorded per channel given the current number of channels added.

        Returns:
            Number of samples that will be recorded per channel with the current channels. Zero if none have been added.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.channel.size",
        )
        response = call(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_buffer_size_async(
            self
    ) -> int:
        """
        Get the number of samples that can be recorded per channel given the current number of channels added.

        Returns:
            Number of samples that will be recorded per channel with the current channels. Zero if none have been added.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="scope.channel.size",
        )
        response = await call_async(
            "oscilloscope/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def start(
            self,
            capture_length: int = 0
    ) -> None:
        """
        Trigger data recording.

        Args:
            capture_length: Optional number of samples to record per channel.
                If left empty, the device records samples until the buffer fills.
                Requires at least Firmware 7.29.
        """
        request = dto.OscilloscopeStartRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            capture_length=capture_length,
        )
        call("oscilloscope/start", request)

    async def start_async(
            self,
            capture_length: int = 0
    ) -> None:
        """
        Trigger data recording.

        Args:
            capture_length: Optional number of samples to record per channel.
                If left empty, the device records samples until the buffer fills.
                Requires at least Firmware 7.29.
        """
        request = dto.OscilloscopeStartRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            capture_length=capture_length,
        )
        await call_async("oscilloscope/start", request)

    def stop(
            self
    ) -> None:
        """
        End data recording if currently in progress.
        """
        request = dto.OscilloscopeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        call("oscilloscope/stop", request)

    async def stop_async(
            self
    ) -> None:
        """
        End data recording if currently in progress.
        """
        request = dto.OscilloscopeRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        await call_async("oscilloscope/stop", request)

    def read(
            self
    ) -> List[OscilloscopeData]:
        """
        Reads the last-recorded data from the oscilloscope. Will block until any in-progress recording completes.

        Returns:
            Array of recorded channel data arrays, in the order added.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call(
            "oscilloscope/read",
            request,
            dto.OscilloscopeReadResponse.from_binary)
        return list(map(OscilloscopeData, response.data_ids))

    async def read_async(
            self
    ) -> List[OscilloscopeData]:
        """
        Reads the last-recorded data from the oscilloscope. Will block until any in-progress recording completes.

        Returns:
            Array of recorded channel data arrays, in the order added.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = await call_async(
            "oscilloscope/read",
            request,
            dto.OscilloscopeReadResponse.from_binary)
        return list(map(OscilloscopeData, response.data_ids))



# This file is generated. Do not modify by hand.
from enum import Enum


class RotationDirection(Enum):
    """
    Direction of rotation.
    """

    CLOCKWISE = 0
    COUNTERCLOCKWISE = 1
    CW = 0
    CCW = 1



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .axis_address import AxisAddress as AxisAddress
from .device_db_source_type import DeviceDbSourceType as DeviceDbSourceType
from .firmware_version import FirmwareVersion as FirmwareVersion
from .log_output_mode import LogOutputMode as LogOutputMode
from .measurement import Measurement as Measurement
from .rotation_direction import RotationDirection as RotationDirection



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import List
from ..call import call, call_async, call_sync

from ..dto import requests as dto
from .axis import Axis
from ..dto.measurement import Measurement
from ..units import LengthUnits


class AxisGroup:
    """
    Groups multiple axes across devices into a single group to allow for simultaneous movement.
    Note that the movement is not coordinated and trajectory is inconsistent and not repeatable between calls.
    Make sure that any possible trajectory is clear of potential obstacles.
    The movement methods return after all the axes finish the movement successfully
    or throw an error as soon as possible.
    """

    @property
    def axes(self) -> List[Axis]:
        """
        Axes of the group.
        """
        return self._axes

    def __init__(self, axes: List[Axis]):
        """
        Initializes the group with the axes to be controlled.
        """
        self._axes = axes.copy()

    def home(
            self
    ) -> None:
        """
        Homes the axes.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        call("axes/home", request)

    async def home_async(
            self
    ) -> None:
        """
        Homes the axes.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        await call_async("axes/home", request)

    def stop(
            self
    ) -> None:
        """
        Stops the axes.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        call("axes/stop", request)

    async def stop_async(
            self
    ) -> None:
        """
        Stops the axes.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        await call_async("axes/stop", request)

    def move_absolute(
            self,
            *position: Measurement
    ) -> None:
        """
        Moves the axes to absolute position.

        Args:
            position: Position.
        """
        request = dto.AxesMoveRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            position=list(position),
        )
        call("axes/move_absolute", request)

    async def move_absolute_async(
            self,
            *position: Measurement
    ) -> None:
        """
        Moves the axes to absolute position.

        Args:
            position: Position.
        """
        request = dto.AxesMoveRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            position=list(position),
        )
        await call_async("axes/move_absolute", request)

    def move_relative(
            self,
            *position: Measurement
    ) -> None:
        """
        Move axes to position relative to the current position.

        Args:
            position: Position.
        """
        request = dto.AxesMoveRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            position=list(position),
        )
        call("axes/move_relative", request)

    async def move_relative_async(
            self,
            *position: Measurement
    ) -> None:
        """
        Move axes to position relative to the current position.

        Args:
            position: Position.
        """
        request = dto.AxesMoveRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            position=list(position),
        )
        await call_async("axes/move_relative", request)

    def move_min(
            self
    ) -> None:
        """
        Moves axes to the minimum position as specified by limit.min.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        call("axes/move_min", request)

    async def move_min_async(
            self
    ) -> None:
        """
        Moves axes to the minimum position as specified by limit.min.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        await call_async("axes/move_min", request)

    def move_max(
            self
    ) -> None:
        """
        Moves axes to the maximum position as specified by limit.max.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        call("axes/move_max", request)

    async def move_max_async(
            self
    ) -> None:
        """
        Moves axes to the maximum position as specified by limit.max.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        await call_async("axes/move_max", request)

    def wait_until_idle(
            self
    ) -> None:
        """
        Waits until all the axes stop moving.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        call("axes/wait_until_idle", request)

    async def wait_until_idle_async(
            self
    ) -> None:
        """
        Waits until all the axes stop moving.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        await call_async("axes/wait_until_idle", request)

    def is_busy(
            self
    ) -> bool:
        """
        Returns bool indicating whether any of the axes is executing a motion command.

        Returns:
            True if any of the axes is currently executing a motion command. False otherwise.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        response = call(
            "axes/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_busy_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether any of the axes is executing a motion command.

        Returns:
            True if any of the axes is currently executing a motion command. False otherwise.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        response = await call_async(
            "axes/is_busy",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def is_homed(
            self
    ) -> bool:
        """
        Returns bool indicating whether all the axes are homed.

        Returns:
            True if all the axes are homed. False otherwise.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        response = call(
            "axes/is_homed",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_homed_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether all the axes are homed.

        Returns:
            True if all the axes are homed. False otherwise.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        response = await call_async(
            "axes/is_homed",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def get_position(
            self,
            *unit: LengthUnits
    ) -> List[float]:
        """
        Returns current axes position.
        The positions are requested sequentially.
        The result position may not be accurate if the axes are moving.

        Args:
            unit: Units of position. You can specify units once or for each axis separately.

        Returns:
            Axes position.
        """
        request = dto.AxesGetSettingRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            setting="pos",
            unit=list(unit),
        )
        response = call(
            "axes/get_setting",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    async def get_position_async(
            self,
            *unit: LengthUnits
    ) -> List[float]:
        """
        Returns current axes position.
        The positions are requested sequentially.
        The result position may not be accurate if the axes are moving.

        Args:
            unit: Units of position. You can specify units once or for each axis separately.

        Returns:
            Axes position.
        """
        request = dto.AxesGetSettingRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
            setting="pos",
            unit=list(unit),
        )
        response = await call_async(
            "axes/get_setting",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the axes.

        Returns:
            A string that represents the axes.
        """
        request = dto.AxesEmptyRequest(
            interfaces=[axis.device.connection.interface_id for axis in self._axes],
            devices=[axis.device.device_address for axis in self._axes],
            axes=[axis.axis_number for axis in self._axes],
        )
        response = call_sync(
            "axes/to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from collections.abc import Iterable
import zaber_bson
from .can_set_state_axis_response import CanSetStateAxisResponse


@dataclass
class CanSetStateDeviceResponse:
    """
    An object containing any setup issues that will prevent setting a state to a given device.
    """

    axis_errors: List[CanSetStateAxisResponse]
    """
    A list of errors that block setting state of device's axes.
    """

    error: Optional[str] = None
    """
    The error blocking applying this state to the given device.
    """

    @staticmethod
    def zero_values() -> 'CanSetStateDeviceResponse':
        return CanSetStateDeviceResponse(
            error=None,
            axis_errors=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'CanSetStateDeviceResponse':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return CanSetStateDeviceResponse.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'error': str(self.error) if self.error is not None else None,
            'axisErrors': [item.to_dict() for item in self.axis_errors] if self.axis_errors is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'CanSetStateDeviceResponse':
        return CanSetStateDeviceResponse(
            error=data.get('error'),  # type: ignore
            axis_errors=[CanSetStateAxisResponse.from_dict(item) for item in data.get('axisErrors')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.error is not None:
            if not isinstance(self.error, str):
                raise ValueError(f'Property "Error" of "CanSetStateDeviceResponse" is not a string.')

        if self.axis_errors is not None:
            if not isinstance(self.axis_errors, Iterable):
                raise ValueError('Property "AxisErrors" of "CanSetStateDeviceResponse" is not iterable.')

            for i, axis_errors_item in enumerate(self.axis_errors):
                if axis_errors_item is None:
                    raise ValueError(f'Item {i} in property "AxisErrors" of "CanSetStateDeviceResponse" is None.')

                if not isinstance(axis_errors_item, CanSetStateAxisResponse):
                    raise ValueError(f'Item {i} in property "AxisErrors" of "CanSetStateDeviceResponse" is not an instance of "CanSetStateAxisResponse".')

                axis_errors_item.validate()



class Int32:
    """
    A signed integer with a 32-bit fixed width.
    """

    def __init__(self, value: int):
        if value < -2 ** 31 or value > 2 ** 31 - 1:
            raise ValueError(f'value {value} cannot be represented in int32')
        self._value = value

    def get_value(self) -> int:
        return self._value

    def __str__(self) -> str:
        return str(self._value)


class Int64:
    """
    A signed integer with a 64-bit fixed width.
    """

    def __init__(self, value: int):
        if value < -2 ** 63 or value > 2 ** 63 - 1:
            raise ValueError(f'value {value} cannot be represented in int64')
        self._value = value

    def get_value(self) -> int:
        return self._value

    def __str__(self) -> str:
        return str(self._value)


class UInt64:
    """
    An unsigned integer with a 64-bit fixed width.
    """

    def __init__(self, value: int):
        if value < 0 or value > 2 ** 64 - 1:
            raise ValueError(f'value {value} cannot be represented in uint64')
        self._value = value

    def get_value(self) -> int:
        return self._value

    def __str__(self) -> str:
        return str(self._value)
